import json

from django.conf import settings
from django.contrib.admin import helpers
from django.contrib.contenttypes.admin import GenericInlineModelAdmin
from django.contrib.staticfiles.storage import staticfiles_storage
try:
    # Django 1.10
    from django.urls import reverse
except ImportError:
    # Django <= 1.9
    from django.core.urlresolvers import reverse
from django import forms
from django.template.defaultfilters import capfirst
from django.utils import six
from django.utils.six.moves import zip
from django.utils.translation import ugettext

from .formsets import NestedInlineFormSet, NestedBaseGenericInlineFormSet
from django.contrib.admin.options import ModelAdmin, InlineModelAdmin


__all__ = (
    'NestedModelAdmin', 'NestedModelAdminMixin', 'NestedInlineAdminFormset',
    'NestedInlineModelAdmin', 'NestedStackedInline', 'NestedTabularInline',
    'NestedInlineModelAdminMixin', 'NestedGenericInlineModelAdmin',
    'NestedGenericStackedInline', 'NestedGenericTabularInline')


def get_method_function(fn):
    return fn.im_func if six.PY2 else fn


class NestedInlineAdminFormset(helpers.InlineAdminFormSet):

    classes = None

    def __init__(self, inline, *args, **kwargs):
        self.request = kwargs.pop('request', None)
        super(NestedInlineAdminFormset, self).__init__(inline, *args, **kwargs)

        if getattr(inline, 'classes', None):
            self.classes = ' '.join(inline.classes)
        else:
            self.classes = ''

    def __iter__(self):
        for inline_admin_form in super(NestedInlineAdminFormset, self).__iter__():
            if not getattr(inline_admin_form.form, 'inlines', None):
                form = inline_admin_form.form
                obj = form.instance if form.instance.pk else None
                formsets, inlines = [], []
                obj_with_nesting_data = form
                if form.prefix.endswith('__prefix__'):
                    obj_with_nesting_data = self.formset
                formsets = getattr(obj_with_nesting_data, 'nested_formsets', None) or []
                inlines = getattr(obj_with_nesting_data, 'nested_inlines', None) or []
                form.inlines = self.model_admin.get_inline_formsets(self.request, formsets, inlines,
                    obj=obj, allow_nested=True)
            for nested_inline in inline_admin_form.form.inlines:
                for nested_form in nested_inline:
                    inline_admin_form.prepopulated_fields += nested_form.prepopulated_fields
            yield inline_admin_form

    def _media(self):
        media = self.opts.media + self.formset.media
        for fs in self:
            media = media + fs.media
            for inline in (getattr(fs.form, 'inlines', None) or []):
                media = media + inline.media

        # Add nested-admin js and css here, to ensure it goes after any widgets
        static_url = staticfiles_storage.url
        server_data_js = reverse('nesting_server_data')
        min_ext = '' if getattr(settings, 'NESTED_ADMIN_DEBUG', False) else '.min'
        return media + forms.Media(
            js=(
                server_data_js,
                static_url('nested_admin/dist/nested_admin%s.js' % min_ext),
            ),
            css={'all': (
                static_url('nested_admin/dist/nested_admin%s.css' % min_ext),
            )})
    media = property(_media)

    @property
    def inline_model_id(self):
        return "-".join([self.opts.opts.app_label, self.opts.opts.model_name])

    def inline_formset_data(self):
        verbose_name = self.opts.verbose_name
        return json.dumps({
            'name': '#%s' % self.formset.prefix,
            'options': {
                'prefix': self.formset.prefix,
                'addText': ugettext('Add another %(verbose_name)s') % {
                    'verbose_name': capfirst(verbose_name),
                },
                'deleteText': ugettext('Remove'),
            },
            'nestedOptions': {
                'sortableFieldName': getattr(self.opts, 'sortable_field_name', None),
                'lookupRelated': getattr(self.opts, 'related_lookup_fields', {}),
                'lookupAutocomplete': getattr(self.opts, 'autocomplete_lookup_fields', {}),
                'formsetFkName': self.formset.fk.name if getattr(self.formset, 'fk', None) else '',
                'nestingLevel': getattr(self.formset, 'nesting_depth', 0),
                'fieldNames': {
                    'position': getattr(self.opts, 'sortable_field_name', None),
                    'pk': self.opts.opts.pk.name,
                },
                'inlineModel': self.inline_model_id,
                'sortableOptions': self.opts.sortable_options,
            },
        })

    @property
    def handler_classes(self):
        classes = set(getattr(self.opts, 'handler_classes', None) or [])
        return tuple(classes | {"djn-model-%s" % self.inline_model_id})


class NestedModelAdminMixin(object):

    inline_admin_formset_helper_cls = NestedInlineAdminFormset

    def get_inline_formsets(self, request, formsets, inline_instances,
                            obj=None, allow_nested=False):
        inline_admin_formsets = []
        for inline, formset in zip(inline_instances, formsets):
            if not allow_nested and getattr(formset, 'is_nested', False):
                continue
            fieldsets = list(inline.get_fieldsets(request, obj))
            readonly = list(inline.get_readonly_fields(request, obj))
            prepopulated = dict(inline.get_prepopulated_fields(request, obj))
            inline_admin_formset = self.inline_admin_formset_helper_cls(
                inline, formset, fieldsets, prepopulated, readonly,
                model_admin=self, request=request)
            inline_admin_formsets.append(inline_admin_formset)
        return inline_admin_formsets

    def _create_formsets(self, request, obj, change):
        orig_formsets, orig_inline_instances = (
            super(NestedModelAdminMixin, self)._create_formsets(
                request, obj, change))

        formsets = []
        inline_instances = []
        prefixes = {}

        for formset, inline_instance in zip(orig_formsets, orig_inline_instances):
            if not hasattr(formset, 'nesting_depth'):
                formset.nesting_depth = 1

            formsets.append(formset)
            inline_instances.append(inline_instance)

            if getattr(inline_instance, 'inlines', []):
                inlines_and_formsets = [
                    (nested, formset)
                    for nested in inline_instance.get_inline_instances(request)]
                i = 0
                while i < len(inlines_and_formsets):
                    nested, formset = inlines_and_formsets[i]
                    i += 1
                    formset_forms = list(formset.forms) + [None]
                    for form in formset_forms:
                        if form is not None:
                            form.parent_formset = formset
                            form_prefix = form.prefix
                            form_obj = form.instance
                        else:
                            form_prefix = formset.add_prefix('empty')
                            form_obj = None
                        InlineFormSet = nested.get_formset(request, form_obj)
                        prefix = '%s-%s' % (form_prefix, InlineFormSet.get_default_prefix())
                        prefixes[prefix] = prefixes.get(prefix, 0) + 1
                        if prefixes[prefix] != 1:
                            prefix = "%s-%s" % (prefix, prefixes[prefix])

                        formset_params = {
                            'instance': form_obj,
                            'prefix': prefix,
                            'queryset': nested.get_queryset(request),
                        }
                        if request.method == 'POST':
                            formset_params.update({
                                'data': request.POST,
                                'files': request.FILES,
                                'save_as_new': '_saveasnew' in request.POST
                            })

                        nested_formset = InlineFormSet(**formset_params)
                        # We set `is_nested` to True so that we have a way
                        # to identify this formset as such and skip it if
                        # there is an error in the POST and we have to create
                        # inline admin formsets.
                        nested_formset.is_nested = True
                        nested_formset.nesting_depth = formset.nesting_depth + 1
                        nested_formset.parent_form = form

                        if form is None:
                            obj = formset
                        else:
                            obj = form
                            if request.method == 'POST':
                                formsets.append(nested_formset)
                                inline_instances.append(nested)
                        obj.nested_formsets = getattr(obj, 'nested_formsets', None) or []
                        obj.nested_inlines = getattr(obj, 'nested_inlines', None) or []
                        obj.nested_formsets.append(nested_formset)
                        obj.nested_inlines.append(nested)

                        if hasattr(nested, 'get_inline_instances'):
                            inlines_and_formsets += [
                                (nested_nested, nested_formset)
                                for nested_nested in nested.get_inline_instances(request)]
        return formsets, inline_instances


class NestedInlineModelAdminMixin(object):

    is_sortable = True
    sortable_field_name = None

    formset = NestedInlineFormSet

    inlines = []

    if 'suit' in settings.INSTALLED_APPS:
        fieldset_template = 'nesting/admin/includes/suit_inline.html'
    elif 'grappelli' in settings.INSTALLED_APPS:
        fieldset_template = 'nesting/admin/includes/grappelli_inline.html'
    else:
        fieldset_template = 'nesting/admin/includes/inline.html'

    def __init__(self, *args, **kwargs):
        sortable_options = {
            'disabled': not(self.is_sortable),
        }
        if hasattr(self, 'sortable_excludes'):
            sortable_options['sortableExcludes'] = self.sortable_excludes
        if hasattr(self, 'sortable_options'):
            sortable_options.update(self.sortable_options)
        self.sortable_options = sortable_options
        super(NestedInlineModelAdminMixin, self).__init__(*args, **kwargs)

    # Copy methods from ModelAdmin
    get_inline_instances = get_method_function(ModelAdmin.get_inline_instances)

    get_formsets_with_inlines = get_method_function(ModelAdmin.get_formsets_with_inlines)

    if hasattr(ModelAdmin, 'get_formsets'):
        get_formsets = get_method_function(ModelAdmin.get_formsets)

    if hasattr(ModelAdmin, '_get_formsets'):
        _get_formsets = get_method_function(ModelAdmin._get_formsets)

    def get_formset(self, request, obj=None, **kwargs):
        FormSet = BaseFormSet = kwargs.pop('formset', self.formset)

        if self.sortable_field_name:
            class FormSet(BaseFormSet):
                sortable_field_name = self.sortable_field_name

        kwargs['formset'] = FormSet
        return super(NestedInlineModelAdminMixin, self).get_formset(request, obj, **kwargs)


class NestedModelAdmin(NestedModelAdminMixin, ModelAdmin):
    pass


class NestedInlineModelAdmin(NestedInlineModelAdminMixin, InlineModelAdmin):
    pass


class NestedStackedInline(NestedInlineModelAdmin):

    if 'grappelli' in settings.INSTALLED_APPS:
        template = 'nesting/admin/inlines/grappelli_stacked.html'
    else:
        template = 'nesting/admin/inlines/stacked.html'


class NestedTabularInline(NestedInlineModelAdmin):

    if 'grappelli' in settings.INSTALLED_APPS:
        template = 'nesting/admin/inlines/grappelli_tabular.html'
        fieldset_template = 'nesting/admin/includes/grappelli_inline_tabular.html'
    else:
        template = 'nesting/admin/inlines/tabular.html'


class NestedGenericInlineModelAdmin(NestedInlineModelAdminMixin, GenericInlineModelAdmin):

    formset = NestedBaseGenericInlineFormSet


class NestedGenericStackedInline(NestedGenericInlineModelAdmin):

    if 'grappelli' in settings.INSTALLED_APPS:
        template = 'nesting/admin/inlines/grappelli_stacked.html'
    else:
        template = 'nesting/admin/inlines/stacked.html'


class NestedGenericTabularInline(NestedGenericInlineModelAdmin):

    if 'grappelli' in settings.INSTALLED_APPS:
        template = 'nesting/admin/inlines/grappelli_tabular.html'
        fieldset_template = 'nesting/admin/includes/grappelli_inline_tabular.html'
    else:
        template = 'nesting/admin/inlines/tabular.html'



import time

from nested_admin.tests.base import BaseNestedAdminTestCase
from .models import GFKRoot, GFKA, GFKB


class TestGenericInlineAdmin(BaseNestedAdminTestCase):

    root_model = GFKRoot

    def test_add_to_empty_one_deep(self):
        root = self.root_model.objects.create(slug='test')

        self.load_admin(root)

        self.add_inline(slug="test")
        self.save_form()

        a_set = root.a_set.all()

        self.assertEqual(len(a_set), 1)
        self.assertEqual(a_set[0].slug, 'test')
        self.assertEqual(a_set[0].position, 0)

    def test_add_to_empty_two_deep(self):
        root = self.root_model.objects.create(slug='test')
        a = GFKA.objects.create(slug='test', content_object=root, position=0)

        self.load_admin(root)
        self.add_inline([0], name="Test")
        self.save_form()

        b_set = a.b_set.all()

        self.assertEqual(len(b_set), 1)
        self.assertEqual(b_set[0].name, "Test")
        self.assertEqual(b_set[0].position, 0)

    def test_drag_existing_objs(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 1],
            screenshot_hack=True)
        self.save_form()

        y_2 = GFKB.objects.get(name='Y 2')

        self.assertEqual(y_2.content_object, x, "item was not moved to the correct parent")
        self.assertEqual(y_2.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/Y 2[1]',
            'root/x[0]/X 1[2]',
            'root/x[0]/X 2[3]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 1[1]'])

    def test_drag_add_drag(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.add_inline(indexes=[0], name='X 3')
        self.drag_and_drop_item(from_indexes=[1, 1], to_indexes=[0, 1],
            screenshot_hack=True)

        self.save_form()

        y_1 = GFKB.objects.get(name='Y 1')
        self.assertEqual(y_1.content_object, x, "Y1 was not moved to the correct parent")
        self.assertEqual(y_1.position, 1, "Y1 was not moved to the correct position")

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/Y 1[1]',
            'root/x[0]/X 1[2]',
            'root/x[0]/X 2[3]',
            'root/x[0]/X 3[4]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 2[1]'])

    def test_drag_new_item(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)

        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)

        self.load_admin(root)

        self.add_inline(indexes=[1], name='Y 2')
        time.sleep(0.01)
        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 1],
            screenshot_hack=True)

        self.save_form()

        y_2 = GFKB.objects.get(name='Y 2')

        self.assertEqual(y_2.content_object, x, "Y2 was not moved to the correct parent")
        self.assertEqual(y_2.position, 1, "Y2 was not moved to the correct position")

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/Y 2[1]',
            'root/x[0]/X 1[2]',
            'root/x[0]/X 2[3]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 1[1]'])

    def test_delete_two_deep(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.delete_inline(indexes=[1, 1])

        self.save_form()

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/X 1[1]',
            'root/x[0]/X 2[2]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 2[1]'])

    def test_delete_one_deep(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.delete_inline(indexes=[0])

        self.save_form()

        self.assertEqual(len(GFKA.objects.filter(slug='x')), 0,
            "GFKA instance was not deleted")

        y = GFKA.objects.get(slug='y')

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[0]/Y 0[0]',
            'root/y[0]/Y 1[1]',
            'root/y[0]/Y 2[2]'])

    def test_delete_two_deep_undelete_one_deep(self):
        """
        Test that, if an item is deleted, then the parent is deleted, and
        then the parent is undeleted, that the item stays deleted.
        """
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.delete_inline(indexes=[0, 1])
        self.delete_inline(indexes=[0])
        self.undelete_inline(indexes=[0])

        self.save_form()

        self.assertEqual(len(GFKA.objects.filter(slug='x')), 1,
            "GFKA instance should not be deleted")

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/X 2[1]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 1[1]',
            'root/y[1]/Y 2[2]'])

    def test_remove_two_deep(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)

        self.load_admin(root)

        self.add_inline(indexes=[1], name='Y 2')
        self.remove_inline(indexes=[1, 2])

        self.save_form()

        self.assertEqual(["%s" % i for i in x.b_set.all()], [
            'root/x[0]/X 0[0]',
            'root/x[0]/X 1[1]',
            'root/x[0]/X 2[2]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 1[1]'])

    def test_drag_item_to_empty_parent(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        y = GFKA.objects.create(slug='y', content_object=root, position=1)
        GFKB.objects.create(name='Y 0', content_object=y, position=0)
        GFKB.objects.create(name='Y 1', content_object=y, position=1)
        GFKB.objects.create(name='Y 2', content_object=y, position=2)

        self.load_admin(root)

        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 0])

        self.save_form()

        y_2 = GFKB.objects.get(name='Y 2')
        self.assertEqual(y_2.content_object, x, "Y2 was not moved to the correct parent")
        self.assertEqual(y_2.position, 0, "Y2 was not moved to the correct position")

        self.assertEqual(["%s" % i for i in x.b_set.all()],
            ['root/x[0]/Y 2[0]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()], [
            'root/y[1]/Y 0[0]',
            'root/y[1]/Y 1[1]'])

    def test_drag_item_to_new_empty_parent(self):
        root = self.root_model.objects.create(slug='root')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        GFKB.objects.create(name='X 0', content_object=x, position=0)
        GFKB.objects.create(name='X 1', content_object=x, position=1)
        GFKB.objects.create(name='X 2', content_object=x, position=2)

        self.load_admin(root)

        self.add_inline(slug="y")
        self.drag_and_drop_item(from_indexes=[0, 2], to_indexes=[1, 0])

        self.save_form()

        x_2 = GFKB.objects.get(name='X 2')
        y = GFKA.objects.get(slug='y')
        self.assertEqual(x_2.content_object, y, "X2 was not moved to the correct parent")
        self.assertEqual(x_2.position, 0, "X2 was not moved to the correct position")

        self.assertEqual(["%s" % i for i in x.b_set.all()],
            ['root/x[0]/X 0[0]', 'root/x[0]/X 1[1]'])

        self.assertEqual(["%s" % i for i in y.b_set.all()],
            ['root/y[1]/X 2[0]'])

    def test_drag_existing_gfkb_to_new_parent_and_back(self):
        root = self.root_model.objects.create(slug='test')
        x = GFKA.objects.create(slug='x', content_object=root, position=0)
        GFKB.objects.create(name='X 0', content_object=x, position=0)

        self.load_admin(root)

        self.add_inline(slug="y")
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 0])

        self.save_form()

        self.assertEqual(len(GFKA.objects.all()), 2, "Save failed")

        x_0 = GFKB.objects.get(name='X 0')

        self.assertEqual(x_0.content_object, x, "X0 is in the wrong parent")
        self.assertEqual(x_0.position, 0, "X0 has the wrong position")



from django.utils import six

from nested_admin.tests.base import BaseNestedAdminTestCase
from .models import Root, AX, AY, BX, BY


class TestIdenticalPrefixes(BaseNestedAdminTestCase):

    root_model = Root

    def test_add_to_empty_one_deep(self):
        root = self.root_model.objects.create(slug='test')
        AX.objects.create(foo='1', root=root, position=0)
        AY.objects.create(foo='2', root=root, position=0)

        self.load_admin(root)
        self.delete_inline([(1, 0)])
        self.save_form()
        ay_objs = AY.objects.all()
        self.assertEqual(len(ay_objs), 0, "AY inline was not deleted")

    def test_drag_and_drop(self):
        root = self.root_model.objects.create(slug='group')
        ax0 = AX.objects.create(foo='x0', root=root, position=0)
        ax1 = AX.objects.create(foo='x1', root=root, position=1)
        ay = AY.objects.create(foo='y0', root=root, position=0)

        BX.objects.create(bar='ax0bx0', a=ax0, position=0)
        BX.objects.create(bar='ax0bx1', a=ax0, position=1)
        BY.objects.create(bar='ax0by0', a=ax0, position=0)
        BY.objects.create(bar='ax0by1', a=ax0, position=1)
        BX.objects.create(bar='ax1bx0', a=ax1, position=0)
        BX.objects.create(bar='ax1bx1', a=ax1, position=1)
        BY.objects.create(bar='ax1by0', a=ax1, position=0)
        BY.objects.create(bar='ax1by1', a=ax1, position=1)
        BX.objects.create(bar='ay0bx0', a=ay, position=0)
        BX.objects.create(bar='ay0bx1', a=ay, position=1)
        BY.objects.create(bar='ay0by0', a=ay, position=0)
        BY.objects.create(bar='ay0by1', a=ay, position=1)

        self.load_admin(root)

        # Move 'ay0by0' to be the first inline in AX[0]/BY[0]
        self.drag_and_drop_item(from_indexes=[(1, 0), (1, 0)], to_indexes=[(0, 0), (1, 0)],
            screenshot_hack=True)

        self.save_form()

        moved_item = BY.objects.get(bar='ay0by0')
        self.assertEqual(six.text_type(moved_item), 'Root(group)/AX[0](x0)/BY[0](ay0by0)',
            "Item was not moved to the correct position")



import os

DEBUG = True
TEMPLATE_DEBUG = DEBUG

SITE_ROOT = os.path.dirname(os.path.realpath(__file__))

ADMIN_MEDIA_PREFIX = '/static/admin/'

MEDIA_ROOT = '%s/media/' % (SITE_ROOT)
MEDIA_URL = '/media/'

STATIC_URL = '/static/'
STATICFILES_DIRS = (
    '%s/static/' % (SITE_ROOT),
)

CACHE_BACKEND = 'dummy:///'
CACHE_KEY_PREFIX = 'semtags'
CACHE_MIDDLEWARE_KEY_PREFIX = CACHE_KEY_PREFIX
CACHE_MIDDLEWARE_SECONDS = 60

LOGIN_REDIRECT_URL = '/'

SITE_NAME = 'semtags.com'

DATABASES = {
    'default': {
        'ENGINE'     : 'django.db.backends.sqlite3',
        'HOST'       : '',
        'PORT'       : '',
        'NAME'       : '%s/semtags.db' % SITE_ROOT,
        'USER'       : '',
        'PASSWORD'   : ''
    }
}

TIME_ZONE = 'America/Chicago'
LANGUAGE_CODE = 'en-us'
USE_I18N = True
SITE_ID = 1
SECRET_KEY = 'gv^gjq&amp;kwrs3uqmd*s-is7%8z7@bc9^#4$txthzx$ta3nrn6(&amp;'
MESSAGE_STORAGE = 'django.contrib.messages.storage.session.SessionStorage'

STATICFILES_FINDERS = (
    'django.contrib.staticfiles.finders.FileSystemFinder',
    'django.contrib.staticfiles.finders.AppDirectoriesFinder',
)

TEMPLATE_LOADERS = (
    'django.template.loaders.filesystem.load_template_source',
    'django.template.loaders.app_directories.load_template_source',
)

TEMPLATE_CONTEXT_PROCESSORS = (
    'django.core.context_processors.auth',
    'django.core.context_processors.debug',
    'django.core.context_processors.media',
    'django.core.context_processors.static',
    'django.contrib.messages.context_processors.messages',
    'django.core.context_processors.request'
)

MIDDLEWARE_CLASSES = (
    'django.middleware.cache.UpdateCacheMiddleware',
    'django.middleware.common.CommonMiddleware',
    'django.middleware.cache.FetchFromCacheMiddleware',
    'django.contrib.sessions.middleware.SessionMiddleware',
    'django.contrib.auth.middleware.AuthenticationMiddleware',
    'django.middleware.csrf.CsrfViewMiddleware',
    'django.contrib.messages.middleware.MessageMiddleware',
)

ROOT_URLCONF = 'example.urls'

TEMPLATE_DIRS = (
    '%s/templates/' % (SITE_ROOT),
)

AUTOCOMPLETE_BACKEND = 'completion.backends.db_backend.DatabaseAutocomplete'
AUTOCOMPLETE_MIN_WORDS = 1

INSTALLED_APPS = (
    'django.contrib.auth',
    'django.contrib.admin',
    'django.contrib.contenttypes',
    'django.contrib.markup',
    'django.contrib.messages',
    'django.contrib.sessions',
    'django.contrib.sites',
    'django.contrib.staticfiles',
    'basic.blog',
    'basic.inlines',
    'basic.media',
    'basic.people',
    'basic.places',
    'tagging', # needed by basic
    'completion',
    'genericm2m',
    
    # lastly, just a single app for this site
    'example.site_app',
)



SECRET_KEY = 'not_empty'
SITE_ID = 1

DATABASES = {
    'default': {
        'ENGINE': 'django.db.backends.sqlite3',
        'NAME': ':memory:',
    }
}

TEMPLATES = [
    {
        'BACKEND': 'django.template.backends.django.DjangoTemplates',
        'DIRS': [],
        'APP_DIRS': True,
        'OPTIONS': {
            'context_processors': [
                'django.template.context_processors.debug',
                'django.template.context_processors.request',
                'django.contrib.auth.context_processors.auth',
                'django.contrib.messages.context_processors.messages'
            ],
        },
    },
]

MIDDLEWARE_CLASSES = tuple()

EMAIL_BACKEND = 'django.core.mail.backends.locmem.EmailBackend'

INSTALLED_APPS = (
    'tests',
)
AUTHENTICATION_BACKENDS = (
    'django.contrib.auth.backends.ModelBackend',
)



from django.contrib import admin
from nested_admin import NestedStackedInline, NestedModelAdmin

from .models import TopLevel, LevelOne, LevelTwo, LevelThree


class LevelThreeInline(NestedStackedInline):
    model = LevelThree
    extra = 1
    inline_classes = ("collapse", "open", "grp-collapse", "grp-open",)
    sortable_field_name = "position"


class LevelTwoInline(NestedStackedInline):
    model = LevelTwo
    extra = 1
    inlines = [LevelThreeInline]
    inline_classes = ("collapse", "open", "grp-collapse", "grp-open",)
    sortable_field_name = "position"


class LevelOneInline(NestedStackedInline):
    model = LevelOne
    extra = 1
    inlines = [LevelTwoInline]
    inline_classes = ("collapse", "open", "grp-collapse", "grp-open",)
    sortable_field_name = "position"


@admin.register(TopLevel)
class TopLevelAdmin(NestedModelAdmin):
    inlines = [LevelOneInline]



import re
from functools import wraps
import json

import django
from django import template
from django.apps import apps
from django.conf import settings
from django.contrib.admin.options import InlineModelAdmin
from django.utils.safestring import mark_safe
from django.utils.html import escape

register = template.Library()


if django.VERSION >= (1, 10):
    from django.templatetags.static import static as _static
else:
    _static = None


django_version = django.VERSION[:2]


@register.simple_tag
def static(path):
    global _static
    if _static is None:
        if apps.is_installed('django-contrib.staticfiles'):
            from django.contrib.staticfiles.templatetags.staticfiles import static as _static
        else:
            from django.templatetags.static import static as _static
    if django.VERSION >= (1, 9) and path == 'admin/img/icon-unknown.gif':
        path = 'admin/img/icon-unknown.svg'
    return _static(path)


@register.filter
def form_index(form):
    """
    The id of the root 'form' element in a formset is the form's 'prefix'
    without the '-' preceding the index of the form. So, for instance, in a case
    where the form's 'id' field has the field name:
        prefix-2-id
    the 'form' element's id would be:
        prefix2
    and the form's index is '2'
    """
    matches = re.search(r'\-(\d+)$', form.prefix)
    if not matches:
        raise Exception("Form with invalid prefix passed to templatetag")
    return int(matches.group(1))


@register.filter
def strip_parent_name(nested_name, parent_name):
    if nested_name.find(parent_name + " ") == 0:
        return nested_name[len(parent_name)+1:]
    else:
        return nested_name
strip_parent_name.is_safe = True

# These tags are defined in grappelli.templatetags.grp_tags. The issue is that
# they are wrapped in mark_safe(), so we can't use them reliably inside of
# attributes.

@register.filter
def json_encode(data):
    return escape(json.dumps(data))

def json_else_list_tag(f):
    """
    Decorator. Registers function as a simple_tag.
    Try: Return value of the decorated function json encoded.
    Except: Return []
    """
    @wraps(f)
    def inner(model_admin):
        try:
            return mark_safe(escape(json.dumps(f(model_admin))))
        except:
            return []
    return register.simple_tag(inner)


@json_else_list_tag
def get_safe_related_lookup_fields_fk(model_admin):
    return model_admin.related_lookup_fields.get("fk", [])


@json_else_list_tag
def get_safe_related_lookup_fields_m2m(model_admin):
    return model_admin.related_lookup_fields.get("m2m", [])


@json_else_list_tag
def get_safe_related_lookup_fields_generic(model_admin):
    return model_admin.related_lookup_fields.get("generic", [])


# AUTOCOMPLETES

@json_else_list_tag
def get_safe_autocomplete_lookup_fields_fk(model_admin):
    return model_admin.autocomplete_lookup_fields.get("fk", [])


@json_else_list_tag
def get_safe_autocomplete_lookup_fields_m2m(model_admin):
    return model_admin.autocomplete_lookup_fields.get("m2m", [])


@json_else_list_tag
def get_safe_autocomplete_lookup_fields_generic(model_admin):
    return model_admin.autocomplete_lookup_fields.get("generic", [])


@register.filter
def formsetsort(formset, arg):
    """
    Takes a list of formset dicts, returns that list sorted by the sortable field.
    """
    if arg:
        sorted_list = []
        for item in formset:
            position = item.form[arg].data
            if position and position != "-1":
                sorted_list.append((int(position), item))
        sorted_list.sort()
        sorted_list = [item[1] for item in sorted_list]
        for item in formset:
            position = item.form[arg].data
            if not position or position == "-1":
                sorted_list.append(item)
    else:
        sorted_list = formset
    return sorted_list


@register.filter
def cell_count(inline_admin_form):
    """Returns the number of cells used in a tabular inline"""
    count = 1  # Hidden cell with hidden 'id' field
    for fieldset in inline_admin_form:
        # Loop through all the fields (one per cell)
        for line in fieldset:
            for field in line:
                count += 1
    if inline_admin_form.formset.can_delete:
        # Delete checkbox
        count += 1
    return count


class IfConditionNode(template.Node):

    def __init__(self, nodelist_true, nodelist_false, value):
        self.nodelist_true = nodelist_true
        self.nodelist_false = nodelist_false
        self.value = value

    def render(self, context):
        if self.value:
            return self.nodelist_true.render(context)
        else:
            return self.nodelist_false.render(context)


def str_to_version(string):
    return tuple([int(s) for s in string.split('.')])


@register.filter
def django_version_lt(string):
    return django_version < str_to_version(string)


@register.filter
def django_version_lte(string):
    return django_version <= str_to_version(string)


@register.filter
def django_version_gt(string):
    return django_version > str_to_version(string)


@register.filter
def django_version_gte(string):
    return django_version >= str_to_version(string)


@register.tag
def ifinlineclasses(parser, token):
    nodelist_true = parser.parse(('else', 'endifinlineclasses'))
    token = parser.next_token()
    if token.contents == 'else':
        nodelist_false = parser.parse(('endifinlineclasses',))
        parser.delete_first_token()
    else:
        nodelist_false = template.NodeList()
    return IfConditionNode(nodelist_true, nodelist_false, hasattr(InlineModelAdmin, 'classes'))


@register.tag
def ifsuit(parser, token):
    nodelist_true = parser.parse(('else', 'endifsuit'))
    token = parser.next_token()
    if token.contents == 'else':
        nodelist_false = parser.parse(('endifsuit',))
        parser.delete_first_token()
    else:
        nodelist_false = template.NodeList()
    return IfConditionNode(nodelist_true, nodelist_false, 'suit' in settings.INSTALLED_APPS)


@register.tag
def ifnotsuit(parser, token):
    nodelist_true = parser.parse(('else', 'endifnotsuit'))
    token = parser.next_token()
    if token.contents == 'else':
        nodelist_false = parser.parse(('endifnotsuit',))
        parser.delete_first_token()
    else:
        nodelist_false = template.NodeList()
    return IfConditionNode(nodelist_true, nodelist_false, 'suit' not in settings.INSTALLED_APPS)



"""
Widgets for Select2 and QuerySetSequence.

They combine :py:class:`~dal_select2.widgets.Select2WidgetMixin` and
:py:class:`~dal_queryset_sequence.QuerySetSequenceSelectMixin` with Django's
Select and SelectMultiple widgets, and are meant to be used with generic model
form fields such as those in :py:mod:`dal_contenttypes`.
"""

from dal_queryset_sequence.widgets import QuerySetSequenceSelectMixin

from dal_select2.widgets import Select2WidgetMixin

from django import forms


class QuerySetSequenceSelect2(Select2WidgetMixin,
                              QuerySetSequenceSelectMixin,
                              forms.Select):
    """Single model select for a generic select2 autocomplete."""


class QuerySetSequenceSelect2Multiple(Select2WidgetMixin,
                                      QuerySetSequenceSelectMixin,
                                      forms.SelectMultiple):
    """Multiple model select for a generic select2 autocomplete."""



import time
from unittest import skipIf, SkipTest

import django

from nested_admin.tests.base import (
    BaseNestedAdminTestCase, expected_failure_if_suit)
from .models import (
    StackedGroup, StackedSection, StackedItem,
    TabularGroup, TabularSection, TabularItem,
    SortableWithExtraRoot, SortableWithExtraChild)


class InlineAdminTestCaseMixin(object):

    @classmethod
    def setUpClass(cls):
        super(InlineAdminTestCaseMixin, cls).setUpClass()
        cls.section_cls, cls.item_cls = cls.nested_models

    def test_add_section_to_empty(self):
        group = self.root_model.objects.create(slug='test')

        self.load_admin(group)

        self.add_inline(slug="test")
        self.save_form()

        sections = group.section_set.all()

        self.assertEqual(len(sections), 1)
        self.assertEqual(sections[0].slug, 'test')
        self.assertEqual(sections[0].position, 0)

    def test_add_item_to_empty(self):
        group = self.root_model.objects.create(slug='test')
        section = self.section_cls.objects.create(slug='test', group=group, position=0)

        self.load_admin(group)

        item_verbose_name = self.item_cls._meta.verbose_name.title()
        with self.clickable_xpath('//a[contains(string(.), "Add another %s")]' % item_verbose_name) as el:
            el.click()
        with self.clickable_xpath('//input[@name="section_set-0-item_set-0-name"]') as el:
            el.send_keys("Test")
        self.save_form()

        items = section.item_set.all()

        self.assertEqual(len(items), 1)
        self.assertEqual(items[0].name, "Test")
        self.assertEqual(items[0].position, 0)

    def test_drag_last_item_between_sections(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)
        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 1],
            screenshot_hack=True)
        self.save_form()

        item_b_2 = self.item_cls.objects.get(name='B 2')
        self.assertEqual(item_b_2.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_2.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/B 2[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    def test_drag_middle_item_between_sections(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.drag_and_drop_item(from_indexes=[1, 1], to_indexes=[0, 1])

        self.save_form()

        item_b_1 = self.item_cls.objects.get(name='B 1')
        self.assertEqual(item_b_1.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_1.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/B 1[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 2[1]'])

    def test_drag_middle_item_between_sections_after_adding_new_item(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.add_inline(indexes=[1], name='B 3')
        self.drag_and_drop_item(from_indexes=[1, 1], to_indexes=[0, 1],
            screenshot_hack=True)

        self.save_form()

        item_b_1 = self.item_cls.objects.get(name='B 1')
        self.assertEqual(item_b_1.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_1.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/B 1[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 2[1]',
            'group/b[1]/B 3[2]'])

    def test_drag_middle_item_between_sections_after_adding_new_item_to_other_section(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.add_inline(indexes=[0], name='A 3')
        self.drag_and_drop_item(from_indexes=[1, 1], to_indexes=[0, 1],
            screenshot_hack=True)

        self.save_form()

        item_b_1 = self.item_cls.objects.get(name='B 1')
        self.assertEqual(item_b_1.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_1.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/B 1[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]',
            'group/a[0]/A 3[4]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 2[1]'])

    def test_drag_new_item_between_sections(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)

        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        self.add_inline(indexes=[1], name='B 2')
        time.sleep(0.01)
        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 1],
            screenshot_hack=True)

        self.save_form()

        item_b_2 = self.item_cls.objects.get(name='B 2')

        self.assertEqual(item_b_2.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_2.position, 1, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/B 2[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    def test_delete_item(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.delete_inline(indexes=[1, 1])

        self.save_form()

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/A 1[1]',
            'group/a[0]/A 2[2]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 2[1]'])

    def test_delete_section(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.delete_inline(indexes=[0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.filter(slug='a')), 0, "Section was not deleted")

        section_b = self.section_cls.objects.get(slug='b')

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[0]/B 0[0]',
            'group/b[0]/B 1[1]',
            'group/b[0]/B 2[2]'])

    @expected_failure_if_suit
    def test_delete_item_undelete_section(self):
        """
        Test that, if an item is deleted, then the section is deleted, and
        then the section is undeleted, that the item stays deleted.
        """
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.delete_inline(indexes=[0, 1])
        self.delete_inline(indexes=[0])
        self.undelete_inline(indexes=[0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.filter(slug='a')), 1, "Section should not be deleted")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/A 2[1]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]',
            'group/b[1]/B 2[2]'])

    def test_remove_item(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        self.add_inline(indexes=[1], name='B 2')
        self.remove_inline(indexes=[1, 2])

        self.save_form()

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/A 1[1]',
            'group/a[0]/A 2[2]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    def test_drag_item_to_empty_section(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 0])

        self.save_form()

        item_b_2 = self.item_cls.objects.get(name='B 2')
        self.assertEqual(item_b_2.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_2.position, 0, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')],
            ['group/a[0]/B 2[0]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    def test_drag_item_to_first_position(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 0],
            screenshot_hack=True)

        self.save_form()

        item_b_2 = self.item_cls.objects.get(name='B 2')
        self.assertEqual(item_b_2.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_2.position, 0, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/B 2[0]',
            'group/a[0]/A 0[1]',
            'group/a[0]/A 1[2]',
            'group/a[0]/A 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    def test_drag_item_to_last_position(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)
        self.item_cls.objects.create(name='B 2', section=section_b, position=2)

        self.load_admin(group)

        self.drag_and_drop_item(from_indexes=[1, 2], to_indexes=[0, 3])

        self.save_form()

        item_b_2 = self.item_cls.objects.get(name='B 2')
        self.assertEqual(item_b_2.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_2.position, 3, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/A 1[1]',
            'group/a[0]/A 2[2]',
            'group/a[0]/B 2[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 0[0]',
            'group/b[1]/B 1[1]'])

    # This test fails with the phantomjs driver on Travis, but it passes locally
    # and it passes with the Chrome driver, so chalking it up to a fluke
    @skipIf(django.VERSION[:2] == (1, 9), "Skipping misbehaving test on travis")
    def test_drag_item_to_new_empty_section(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='A 2', section=section_a, position=2)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.drag_and_drop_item(from_indexes=[0, 2], to_indexes=[1, 0])

        self.save_form()

        item_a_2 = self.item_cls.objects.get(name='A 2')
        section_b = self.section_cls.objects.get(slug='b')
        self.assertEqual(item_a_2.section, section_b, "item was not moved to the correct section")
        self.assertEqual(item_a_2.position, 0, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')],
            ['group/a[0]/A 0[0]', 'group/a[0]/A 1[1]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')],
            ['group/b[1]/A 2[0]'])

    def test_position_update_bug(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)

        self.item_cls.objects.create(name='B 0', section=section_b, position=0)

        self.load_admin(group)

        self.add_inline(indexes=[0], name='A 0')
        self.add_inline(indexes=[0], name='A 1')
        self.add_inline(indexes=[0], name='A 2')

        # Move to second position of the first section
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 1])

        # Move to the last position of the first section
        self.drag_and_drop_item(from_indexes=[0, 1], to_indexes=[0, 3])

        position_selector = self.get_form_field_selector('position', indexes=[0, 3])

        def check_position_is_correct(d):
            val = d.execute_script('return $("%s").val()' % position_selector)
            return val == '3'

        self.wait_until(
            check_position_is_correct,
            message="Timeout waiting for position to update to correct value")

        self.save_form()

        item_b_0 = self.item_cls.objects.get(name='B 0')
        self.assertEqual(item_b_0.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_0.position, 3, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 0[0]',
            'group/a[0]/A 1[1]',
            'group/a[0]/A 2[2]',
            'group/a[0]/B 0[3]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [])

    def test_drag_existing_item_to_new_section_and_back(self):
        group = self.root_model.objects.create(slug='test')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.all()), 2, "Save failed")

        item_a_0 = self.item_cls.objects.get(name='A 0')

        self.assertEqual(item_a_0.section, section_a, "Item is in the wrong section")
        self.assertEqual(item_a_0.position, 0, "Item has the wrong position")

    def test_drag_item_create_invalid_new_item_then_drag_back_after_validation_error_removing_invalid_item(self):
        """
        Tests regression of a scenario after encountering a validation error.

        Steps to reproduce:
            1. Begin with at least two items in each section
            2. Drag one of the items from the first section into the second
            3. Create an invalid item in the first section
            4. Save, encounter a validation error
            5. Drag the invalid item back to the first group
            6. Remove the invalid item
            7. Save, get a 500 Internal Server Error
        """
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        # Drag the first item of section 'b' into section 'a'
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 1])
        # Create invalid item (missing required field 'name')
        self.add_inline(indexes=[1])

        # We need to manually set the position to trigger the validation error,
        # otherwise it will be skipped as an empty inline on save
        self.set_field('position', '1', indexes=[1, 1])

        # Save
        self.save_form()

        self.drag_and_drop_item(from_indexes=[1, 1], to_indexes=[0, 0],
            screenshot_hack=True)
        # Remove invalid item
        self.remove_inline(indexes=[0, 0])
        # Make a change to test whether save succeeds
        self.set_field("name", 'A 0_changed', indexes=[0, 0])

        self.save_form()

        item_a_0 = self.item_cls.objects.get(section=section_a, position=0)
        self.assertEqual(item_a_0.name, 'A 0_changed', 'Save failed')

    def test_swap_first_two_items_between_sections(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 0],
            screenshot_hack=True)
        self.drag_and_drop_item(from_indexes=[0, 1], to_indexes=[1, 0],
            screenshot_hack=True)

        self.save_form()

        item_b_0 = self.item_cls.objects.get(name='B 0')
        self.assertEqual(item_b_0.section, section_a, "item was not moved to the correct section")
        self.assertEqual(item_b_0.position, 0, "item was not moved to the correct position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/B 0[0]',
            'group/a[0]/A 1[1]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/A 0[0]',
            'group/b[1]/B 1[1]'])

    def test_drag_first_item_to_new_section(self):
        """
        Test dragging the first of several items in a pre-existing section into
        a newly created section.
        """
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.all()), 2, "Save failed")

        section_b = self.section_cls.objects.get(slug='b')
        item_a_0 = self.item_cls.objects.get(name='A 0')

        self.assertEqual(item_a_0.section, section_b, "Item is in the wrong section")
        self.assertEqual(item_a_0.position, 0, "Item has the wrong position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 1[0]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/A 0[0]'])

    def test_drag_first_item_to_new_section_after_removing_item(self):
        """
        Test dragging the first of several items in a pre-existing section into
        a newly created section after having added two items and then removing
        one of those items.
        """
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='A 1', section=section_a, position=1)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.add_inline(indexes=[1], name='B 0')
        self.add_inline(indexes=[1], name='B 1')
        self.remove_inline(indexes=[1, 0])
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.all()), 2, "Save failed")

        section_b = self.section_cls.objects.get(slug='b')
        item_a_0 = self.item_cls.objects.get(name='A 0')
        item_a_1 = self.item_cls.objects.get(name='A 1')
        item_b_1 = self.item_cls.objects.get(name='B 1')

        self.assertNotEqual(item_a_0.section, section_a, "A0 did not move to new section")
        self.assertEqual(item_a_0.position, 0, "A0 has the wrong position")
        self.assertEqual(item_a_1.position, 0, "A1 has the wrong position")
        self.assertEqual(item_b_1.position, 1, "B1 has the wrong position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/A 1[0]'])

        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/A 0[0]', 'group/b[1]/B 1[1]'])

    def test_add_remove_items_in_new_section_dragging_existing_items(self):
        """
        Tests for a regression that could be reproduced with the following steps:

        1. Begin with one section, with at least one item in it.
        2. Create a new section
        3. Create three items in the new section
        4. Remove the first of the new items
        5. Drag the first of the existing items into the first position in the
           new section.
        6. Remove the second item in the new section
        8. Save

        Expected outcome:
            The dragged item from the existing section should have been moved
            to the new section.

        Outcome with bug:
            The item has not moved.
        """
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.add_inline(indexes=[1], name='B 0')
        self.add_inline(indexes=[1], name='B 1')
        self.add_inline(indexes=[1], name='B 2')
        self.remove_inline(indexes=[1, 0])
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])
        self.remove_inline(indexes=[1, 1])

        self.save_form()

        section_b = self.section_cls.objects.get(slug='b')
        item_a_0 = self.item_cls.objects.get(name='A 0')
        item_b_2 = self.item_cls.objects.get(name='B 2')

        self.assertNotEqual(item_a_0.section, section_a, "A0 did not move to new section")
        self.assertEqual(item_a_0.position, 0, "A0 has the wrong position")
        self.assertEqual(item_b_2.position, 1, "B2 has the wrong position")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [])
        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/A 0[0]', 'group/b[1]/B 2[1]'])

    def test_delete_section_after_dragging_item_away(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        # Drag the first item of section 'b' into section 'a'
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 0])

        # Delete section 'b'
        self.delete_inline(indexes=[1])

        self.save_form()

        self.assertNotEqual(len(self.section_cls.objects.all()), 2, "Save failed")

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/B 0[0]', 'group/a[0]/A 0[1]'])

    def test_delete_undelete_section_after_dragging_item_away(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        section_b = self.section_cls.objects.create(slug='b', group=group, position=1)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)
        self.item_cls.objects.create(name='B 0', section=section_b, position=0)
        self.item_cls.objects.create(name='B 1', section=section_b, position=1)

        self.load_admin(group)

        # Drag the first item of section 'b' into section 'a'
        self.drag_and_drop_item(from_indexes=[1, 0], to_indexes=[0, 0])

        # Delete section 'b'
        self.delete_inline(indexes=[1])
        self.undelete_inline(indexes=[1])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.all()), 2)

        self.assertEqual(["%s" % i for i in section_a.item_set.all().order_by('position')], [
            'group/a[0]/B 0[0]', 'group/a[0]/A 0[1]'])
        self.assertEqual(["%s" % i for i in section_b.item_set.all().order_by('position')], [
            'group/b[1]/B 1[0]'])

    def test_drag_into_new_section_after_adding_and_removing_preceding_section(self):
        group = self.root_model.objects.create(slug='group')
        section_a = self.section_cls.objects.create(slug='a', group=group, position=0)
        self.item_cls.objects.create(name='A 0', section=section_a, position=0)

        self.load_admin(group)

        self.add_inline(slug="b")
        self.add_inline(slug="c")
        self.remove_inline(indexes=[1])
        self.drag_and_drop_item(from_indexes=[0, 0], to_indexes=[1, 0])

        self.save_form()

        self.assertEqual(len(self.section_cls.objects.all()), 2, "Save failed (new section wasn't added)")

        item_a0 = self.item_cls.objects.get(name='A 0')
        section_c = self.section_cls.objects.get(slug='c')

        self.assertEqual(item_a0.section, section_c, "Item was not moved to new section")


class TestStackedInlineAdmin(InlineAdminTestCaseMixin, BaseNestedAdminTestCase):

    root_model = StackedGroup
    nested_models = (StackedSection, StackedItem)

    def test_add_item_inline_label_update(self):
        if django.VERSION < (1, 9):
            raise SkipTest("Test only applies to Django 1.9+")
        if self.has_grappelli:
            raise SkipTest("Test does not apply if using django-grappelli")
        if self.has_suit:
            raise SkipTest("Test does not apply if using django-suit")
        group = self.root_model.objects.create(slug='test')
        self.section_cls.objects.create(slug='test', group=group, position=0)

        self.load_admin(group)
        item_verbose_name = self.item_cls._meta.verbose_name.title()
        with self.clickable_xpath('//a[contains(string(.), "Add another %s")]' % item_verbose_name) as el:
            el.click()
        with self.clickable_xpath('//input[@name="section_set-0-item_set-0-name"]') as el:
            el.send_keys("Test 1")
        with self.clickable_xpath('//a[contains(string(.), "Add another %s")]' % item_verbose_name) as el:
            el.click()
        with self.clickable_xpath('//input[@name="section_set-0-item_set-0-name"]') as el:
            el.send_keys("Test 2")

        inline_label = self.get_item([0, 1]).find_element_by_class_name('inline_label')
        self.assertEqual(inline_label.text, '#2')


class TestTabularInlineAdmin(InlineAdminTestCaseMixin, BaseNestedAdminTestCase):

    root_model = TabularGroup
    nested_models = (TabularSection, TabularItem)


class TestSortablesWithExtra(BaseNestedAdminTestCase):

    root_model = SortableWithExtraRoot
    nested_models = (SortableWithExtraChild, )

    def test_blank_extra_inlines_validation(self):
        root = self.root_model.objects.create(slug='a')
        self.load_admin(root)
        self.set_field('slug', 'b')

        self.save_form()

        validation_errors = self.selenium.execute_script(
            "return $('ul.errorlist li').length")

        self.assertEqual(validation_errors, 0, "Unexpected validation errors encountered")

    def test_blank_extra_inlines_validation_with_change(self):
        root = self.root_model.objects.create(slug='a')
        self.load_admin(root)
        self.set_field('slug', 'b')
        self.set_field('slug', 'a', indexes=[0])

        self.save_form()

        validation_errors = self.selenium.execute_script(
            "return $('ul.errorlist li').length")

        self.assertEqual(validation_errors, 0, "Unexpected validation errors encountered")

        # refetch from the database
        root = self.root_model.objects.get(pk=root.pk)

        self.assertEqual(root.slug, 'b', "Root slug did not change")

        children = self.nested_models[0].objects.all()

        self.assertNotEqual(len(children), 0, "Child object did not save")
        self.assertEqual(len(children), 1, "Incorrect number of children saved")
        self.assertEqual(children[0].slug, "a", "Child slug incorrect")
